package org.vraptor.core;

import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import javax.servlet.ServletException;

import org.apache.log4j.Logger;
import org.vraptor.Interceptor;
import org.vraptor.LogicException;
import org.vraptor.LogicFlow;
import org.vraptor.LogicRequest;
import org.vraptor.RedirectException;
import org.vraptor.component.ComponentInstantiationException;
import org.vraptor.interceptor.InterceptorDealer;
import org.vraptor.introspector.Introspector;
import org.vraptor.reflection.GettingException;
import org.vraptor.reflection.MethodInvocationException;
import org.vraptor.reflection.SettingException;
import org.vraptor.view.ViewException;

/**
 * The base implementation of logic flow, calling the given interceptors.
 * 
 * @author Guilherme Silveira
 * @author Paulo Silveira
 * 
 */
public class InterceptorsLogicFlow implements LogicFlow {

	private static final Logger LOG = Logger.getLogger(InterceptorsLogicFlow.class);

	private final InterceptorDealer interceptorDealer;

	private final LinkedList<Interceptor> interceptors;

	private Interceptor currentInterceptor;

	private final LogicRequest request;

	public InterceptorsLogicFlow(LogicRequest request, Introspector introspector, List<Interceptor> interceptors) {
		this.request = request;
		this.interceptorDealer = new InterceptorDealer(introspector);
		this.interceptors = new LinkedList<Interceptor>(interceptors);
	}

	/**
	 * Executes the next step in the flow, keeping in mind the last interceptor
	 * executed, so we can execute the next one.
	 * 
	 */
	public void execute() throws ViewException, LogicException {
		if (currentInterceptor != null) {
			try {
				interceptorDealer.outject(currentInterceptor, request);
			} catch (GettingException e) {
				throw new LogicException(e.getMessage(), e);
			} catch (MethodInvocationException e) {
				throw new LogicException(e);
			}
		}

		if (interceptors.isEmpty()) {
			throw new IllegalStateException("logicFlow.execute being called more than"
					+ " than the number of interceptors. The last interceptor" + "must not continue the flow!");
		}
		currentInterceptor = interceptors.poll();

		try {
			interceptorDealer.inject(currentInterceptor, request);
		} catch (ComponentInstantiationException e) {
			throw new LogicException(e.getMessage(), e);
		} catch (SettingException e) {
			throw new LogicException(e.getMessage(), e);
		}

		if (LOG.isDebugEnabled()) {
			//TODO trace
			LOG.debug("Calling interceptor: " + currentInterceptor);
		}
		currentInterceptor.intercept(this);
	}

	public void addInterceptor(Interceptor interceptor) {
		this.interceptors.addFirst(interceptor);
	}

	public void redirect(String url) throws RedirectException {
		LOG.warn("redirecting using a deprecated method. See LogicFlow.redirect");
		try {
			request.getRequest().getRequestDispatcher(url).forward(request.getRequest(), request.getResponse());
		} catch (ServletException e) {
			if (e.getRootCause() == null) {
				throw new RedirectException(e.getMessage(), e);
			}
			throw new RedirectException(e.getRootCause().getMessage(), e.getRootCause());
		} catch (IOException e) {
			throw new RedirectException(e.getMessage(), e);
		}
	}

	public LogicRequest getLogicRequest() {
		return this.request;
	}

}
