package org.vraptor.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.apache.log4j.Logger;
import org.vraptor.Interceptor;
import org.vraptor.LogicException;
import org.vraptor.LogicRequest;
import org.vraptor.component.ComponentInstantiationException;
import org.vraptor.component.ComponentNotFoundException;
import org.vraptor.component.ComponentType;
import org.vraptor.component.LogicMethod;
import org.vraptor.component.LogicNotFoundException;
import org.vraptor.http.VRaptorServletRequest;
import org.vraptor.interceptor.ComponentLookupInterceptor;
import org.vraptor.interceptor.ExecuteLogicInterceptor;
import org.vraptor.interceptor.InjectionInterceptor;
import org.vraptor.interceptor.InterceptorInstantiationException;
import org.vraptor.interceptor.InterceptorType;
import org.vraptor.interceptor.OutjectionInterceptor;
import org.vraptor.interceptor.RegisterAttributesInteceptor;
import org.vraptor.interceptor.SettingAndValidationInterceptor;
import org.vraptor.interceptor.ViewInterceptor;
import org.vraptor.scope.DefaultLogicRequest;
import org.vraptor.url.RequestInfo;
import org.vraptor.view.ViewException;

/**
 * VRaptor's engine.
 * 
 * @author Guilherme Silveira
 * @author Paulo Silveira
 */
public class VRaptorExecution {

	private static final Logger LOG = Logger.getLogger(VRaptorExecution.class);

	private final VRaptorController controller;

	private final RequestInfo requestInfo;

	private final WebRequest webRequest;

	public VRaptorExecution(RequestInfo requestInfo, VRaptorController controller, WebRequest webRequest) {
		this.webRequest = webRequest;
		this.controller = controller;
		this.requestInfo = requestInfo;
	}

	public String execute() throws ComponentNotFoundException, LogicNotFoundException, ViewException,
			InterceptorInstantiationException, LogicException {
		if (LOG.isDebugEnabled()) {
			LOG.debug("Calling execute on " + requestInfo.toString());
		}

		ComponentType componentType = this.controller.getWebApplication().getComponentManager().getComponent(
				requestInfo);
		LogicMethod method = componentType.getLogic(requestInfo.getLogicName());

		VRaptorServletRequest vraptorRequest = new VRaptorServletRequest(webRequest.getRequest(), controller);

		// wrapping VRaptorServletRequest
		WebRequest wrappedWebRequest = new WebRequest(vraptorRequest, webRequest.getResponse(), webRequest.getContext());
		LogicRequest logicRequest = new DefaultLogicRequest(requestInfo, wrappedWebRequest, componentType, method);

		// TODO: are these necessary? at least externalize
		// FIXME: vraptorRequest and logcRequest have a bidirection dependency
		vraptorRequest.setCurrentRequest(logicRequest);
		vraptorRequest.setAttribute("context", logicRequest);
		vraptorRequest.setAttribute("application", controller.getWebApplication().getApplicationContext());

		Interceptor register = new RegisterAttributesInteceptor(method.getMetadata(), componentType);

		List<Interceptor> after = new ArrayList<Interceptor>();

		after.add(new ComponentLookupInterceptor(controller.getWebApplication().getIntrospector().getBeanProvider()));
		after.add(new InjectionInterceptor(controller.getIntrospector(), componentType));
		after.add(new SettingAndValidationInterceptor(componentType, controller.getIntrospector(), controller
				.getWebApplication().getConverterManager(), controller.getWebApplication().getViewManager(), method));
		after.add(new ExecuteLogicInterceptor());
		after.add(new OutjectionInterceptor(controller.getIntrospector(), componentType));
		after.add(new ViewInterceptor(controller.getWebApplication().getViewManager(), method));

		List<Interceptor> interceptors = generateInterceptors(Collections.singletonList(register), componentType, after);

		InterceptorsLogicFlow logicFlow = new InterceptorsLogicFlow(logicRequest, controller.getIntrospector(),
				interceptors);

		logicFlow.execute();

		return logicFlow.getLogicRequest().getResult();
	}

	/**
	 * stacks the interceptors into a single list
	 * 
	 * @param before
	 * @param componentType
	 * @param after
	 * @return
	 * @throws InterceptorInstantiationException
	 */
	private static List<Interceptor> generateInterceptors(List<Interceptor> before, ComponentType componentType,
			List<Interceptor> after) throws InterceptorInstantiationException {
		List<Interceptor> interceptors = new ArrayList<Interceptor>();
		interceptors.addAll(before);
		for (InterceptorType type : componentType.getInterceptors()) {
			try {
				Interceptor interceptor = type.newInstance();
				if (LOG.isDebugEnabled()) {
					LOG.debug("Adding interceptor " + interceptor.getClass().getName());
				}
				interceptors.add(interceptor);
			} catch (ComponentInstantiationException e) {
				throw new InterceptorInstantiationException(e.getMessage(), e);
			}
		}
		interceptors.addAll(after);
		return interceptors;
	}

}
